@@ -17,8 +17,6 @@ from firebirdsql.wireprotocol import (WireProtocol,
     bytes_to_bint, bytes_to_int, bint_to_bytes, int_to_bytes, 
     INFO_SQL_SELECT_DESCRIBE_VARS,)
 
-PYTHON_MAJOR_VER = sys.version_info[0]
-
 if sys.version_info < (2, 6):
     from UserDict import DictMixin as Mapping
     HAS_MAPPING = False
@@ -26,20 +24,34 @@ else:
     from collections import Mapping
     HAS_MAPPING = True
 
+try:
+    import fcntl
+except ImportError:
+    def setcloexec(sock):
+        pass
+else:
+    def setcloexec(sock):
+        """Set FD_CLOEXEC property on a file descriptors
+        """
+        fd = sock.fileno()
+        flags = fcntl.fcntl(fd, fcntl.F_GETFD)
+        fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
+
+
 def b2i(b):
     "byte to int"
-    if PYTHON_MAJOR_VER == 3:
+    if sys.version_info[0] == 3:
         return b
     else:
         return ord(b)
 
-if PYTHON_MAJOR_VER == 2:
+if sys.version_info[0] == 2:
     __metaclass__ = type
 
     def bytes(byte_array):
         return ''.join(chr(c) for c in byte_array)
 
-__version__ = '0.6.3'
+__version__ = '0.6.4'
 apilevel = '2.0'
 threadsafety = 1
 paramstyle = 'qmark'
@@ -81,7 +93,7 @@ class DBAPITypeObject:
         else:
             return -1
 STRING = DBAPITypeObject(str)
-if PYTHON_MAJOR_VER==3:
+if sys.version_info[0]==3:
     BINARY = DBAPITypeObject(bytes)
 else:
     BINARY = DBAPITypeObject(str)
@@ -122,6 +134,21 @@ class XSQLVAR:
         SQL_TYPE_INT64: 8
         }
     
+    type_display_length = {
+        SQL_TYPE_VARYING: -1,
+        SQL_TYPE_SHORT: 6,
+        SQL_TYPE_LONG: 11,
+        SQL_TYPE_FLOAT: 17,
+        SQL_TYPE_TIME: 11,
+        SQL_TYPE_DATE: 10,
+        SQL_TYPE_DOUBLE: 17,
+        SQL_TYPE_TIMESTAMP: 22,
+        SQL_TYPE_BLOB: 0,
+        SQL_TYPE_ARRAY: -1,
+        SQL_TYPE_QUAD: 20,
+        SQL_TYPE_INT64: 20
+        }
+
     def __init__(self, bytes_to_str):
         self.bytes_to_str = bytes_to_str
         self.sqltype = None
@@ -141,6 +168,16 @@ class XSQLVAR:
         else:
             return self.type_length[sqltype]
 
+    def display_length(self):
+        sqltype = self.sqltype
+        if sqltype == SQL_TYPE_TEXT:
+            return self.sqllen
+        else:
+            return self.type_display_length[sqltype]
+
+    def precision(self):
+        return None
+
     def __str__(self):
         s  = '[' + str(self.sqltype) + ',' + str(self.sqlscale) + ',' \
                 + str(self.sqlsubtype) + ',' + str(self.sqllen)  + ',' \
@@ -313,7 +350,8 @@ def parse_select_items(buf, xsqlda, connection):
 def parse_xsqlda(buf, connection, stmt_handle):
     assert buf[:3] == bytes([0x15,0x04,0x00]) # isc_info_sql_stmt_type
     stmt_type = bytes_to_int(buf[3:7])
-    if stmt_type != isc_info_sql_stmt_select:
+    if (stmt_type != isc_info_sql_stmt_select and
+        stmt_type != isc_info_sql_stmt_exec_procedure):
         return []
 
     assert buf[7:9] == bytes([0x04,0x07])
@@ -365,7 +403,8 @@ class PreparedStatement:
                 return None
             r = []
             for x in self._xsqlda:
-                r.append((x.aliasname, x.sqltype, None, x.io_length(), None, 
+                r.append((x.aliasname, x.sqltype, x.diplay_length(), 
+                        x.io_length(), x.precision(), 
                         x.sqlscale, True if x.null_ok else False))
             return r
         elif attrname == 'n_output_params':
@@ -384,9 +423,6 @@ class Cursor:
     def transaction(self):
         return self._transaction
 
-    def callproc(self, procname, *params):
-        raise NotSupportedError()
-
     def _convert_params(self, params):
         cooked_params = []
         for param in params:        # Convert str/bytes parameter to blob id
@@ -394,7 +430,7 @@ class Cursor:
                 param = self.transaction.connection.str_to_bytes(param)
             cooked_params.append(param)
             continue
-            self.transaction.connection._op_create_blob2()
+            self.transaction.connection._op_create_blob2(self.transaction.trans_handle)
             (blob_handle, blob_id, buf2) = self.transaction.connection._op_response()
             seg_size = self.transaction.connection.buffer_length
             (seg, remains) = param[:seg_size], param[seg_size:]
@@ -435,6 +471,22 @@ class Cursor:
                 raise IntegrityError(e.message, e.gds_codes, e.sql_code)
         return stmt_type, stmt_handle
 
+    def _callproc(self, query, params):
+        cooked_params = self._convert_params(params)
+        stmt_handle = self.stmt_handle
+        self.transaction.connection._op_prepare_statement(stmt_handle, 
+                                        self.transaction.trans_handle, query)
+        (h, oid, buf) = self.transaction.connection._op_response()
+        assert buf[:3] == bytes([0x15,0x04,0x00]) # isc_info_sql_stmt_type
+        stmt_type = bytes_to_int(buf[3:7])
+        self._xsqlda = parse_xsqlda(buf, self.transaction.connection, 
+                                                                stmt_handle)
+
+        self.transaction.connection._op_execute2(stmt_handle,
+            self.transaction.trans_handle, cooked_params,
+            calc_blr(self._xsqlda))
+        return self.transaction.connection._op_sql_response(self._xsqlda)
+
     def prep(self, query):
         prepared_statement = PreparedStatement(self, query)
         return prepared_statement
@@ -444,6 +496,10 @@ class Cursor:
         if stmt_type == isc_info_sql_stmt_select:
             self._fetch_records = self._fetch_generator(stmt_handle)
 
+    def callproc(self, procname, params = []):
+        query = 'EXECUTE PROCEDURE ' + procname + ' ' + ','.join('?'*len(params))
+        return self._callproc(query, params)
+
     def executemany(self, query, seq_of_params):
         for params in seq_of_params:
             self.execute(query, params)
@@ -462,7 +518,7 @@ class Cursor:
                     if x.sqltype == SQL_TYPE_BLOB:    
                         if not r[i]:
                             continue
-                        connection._op_open_blob(r[i])
+                        connection._op_open_blob(r[i], self.transaction.trans_handle)
                         (h, oid, buf) = connection._op_response()
                         v = bytes([])
                         n = 1   # 1:mora data 2:no more data
@@ -489,7 +545,7 @@ class Cursor:
 
     def fetchone(self):
         try:
-            if PYTHON_MAJOR_VER==3:
+            if sys.version_info[0] == 3:
                 return next(self._fetch_records)
             else:
                 return self._fetch_records.next()
@@ -546,8 +602,8 @@ class Cursor:
     
     @property
     def description(self):
-        return [(x.aliasname, x.sqltype, None, x.io_length(), None, 
-                 x.sqlscale, True if x.null_ok else False)
+        return [(x.aliasname, x.sqltype, x.display_length(), x.io_length(), 
+                    x.precision(), x.sqlscale, True if x.null_ok else False)
                 for x in self._xsqlda]
         
     @property
@@ -563,59 +619,10 @@ class Connection(WireProtocol):
             hostname = os.environ['COMPUTERNAME']
         else:
             user = os.environ['USER']
-            hostname = os.environ.get('HOSTNAME', '')
+            hostname = socket.gethostname()
         return bytes([1] + [len(user)] + [ord(c) for c in user] 
                 + [4] + [len(hostname)] + [ord(c) for c in hostname] + [6, 0])
 
-    def params_to_blr(self, params):
-        "Convert parameter array to BLR and values format."
-        ln = len(params) * 2
-        blr = bytes([5, 2, 4, 0, ln & 255, ln >> 8])
-        values = bytes([])
-        for p in params:
-            t = type(p)
-            if ((PYTHON_MAJOR_VER == 2 and t == str) or
-                (PYTHON_MAJOR_VER == 3 and t == bytes)):
-                v = p
-                nbytes = len(v)
-                pad_length = ((4-nbytes) & 3)
-                v += bytes([0]) * pad_length
-                blr += bytes([14, nbytes & 255, nbytes >> 8])
-            elif t == int:
-                v = bint_to_bytes(p, 4)
-                blr += bytes([7, 0])
-            elif t == decimal.Decimal or t == float:
-                if t == float:
-                    p = decimal.Decimal(str(p))
-                (sign, digits, exponent) = p.as_tuple()
-                v = 0
-                ln = len(digits)
-                for i in range(ln):
-                    v += digits[i] * (10 ** (ln -i-1))
-                if sign:
-                    v *= -1
-                v = bint_to_bytes(v, 8)
-                if exponent < 0:
-                    exponent += 256
-                blr += bytes([16, exponent])
-            elif t == datetime.date:
-                v = convert_date(p)
-                blr += bytes([12])
-            elif t == datetime.time:
-                v = convert_time(p)
-                blr += bytes([13])
-            elif t == datetime.datetime:
-                v = convert_timestamp(p)
-                blr += bytes([35])
-            elif p == None:
-                v = bytes([0]) * 8
-                blr += bytes([9, 0])
-            values += v
-            blr += bytes([7, 0])
-            values += bytes([0]) * 4 if p != None else bytes([0xff,0xff,0x34,0x8c])
-        blr += bytes([255, 76])    # [blr_end, blr_eoc]
-        return blr, values
-
     def cursor(self):
         if self.main_transaction is None:
             self.begin()
@@ -651,8 +658,7 @@ class Connection(WireProtocol):
 
     def __init__(self, dsn=None, user=None, password=None, host=None,
                     database=None, charset=DEFAULT_CHARSET, port=3050, 
-                    page_size=None,
-                    is_services = False):
+                    page_size=None, is_services = False, cloexec=False):
         if dsn:
             i = dsn.find(':')
             if i < 0:
@@ -664,14 +670,15 @@ class Connection(WireProtocol):
                 i = hostport.find('/')
                 if i < 0:
                     self.hostname = hostport
-                    self.port = port
                 else:
                     self.hostname = hostport[:i]
-                    self.port = int(hostport[i+1:])
+                    port = int(hostport[i+1:])
         else:
             self.hostname = host
             self.filename = database
-            self.port = port
+        if self.hostname is None:
+            self.hostname = 'localhost'
+        self.port = port
         self.user = user
         self.password = password
         self.charset = charset
@@ -679,6 +686,8 @@ class Connection(WireProtocol):
         self.isolation_level = ISOLATION_LEVEL_READ_COMMITED
 
         self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+        if cloexec:
+            setcloexec(self.sock)
         self.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
         self.sock.connect((self.hostname, self.port))
 
@@ -827,6 +836,11 @@ class Connection(WireProtocol):
                                                     info_requests[i], rs[i])
             return results
 
+    def trans_info(self, info_requests):
+        if self.main_transaction:
+            return self.main_transaction.trans_info(info_requests)
+        return {}
+
     def close(self):
         if self.closed:
             return
@@ -892,6 +906,43 @@ class Transaction:
             delattr(self, "trans_handle")
             self.connection._transactions.remove(self)
 
+    def _trans_info(self, info_requests):
+        if info_requests[-1] == isc_info_end:
+            self.connection._op_info_transaction(self.trans_handle,
+                    bytes(info_requests))
+        else:
+            self.connection._op_info_transaction(self.trans_handle,
+                    bytes(info_requests+type(info_requests)([isc_info_end])))
+        (h, oid, buf) = self.connection._op_response()
+        i = 0
+        i_request = 0
+        r = []
+        while i < len(buf):
+            req = b2i(buf[i])
+            if req == isc_info_end:
+                break
+            assert req == info_requests[i_request]
+            l = bytes_to_int(buf[i+1:i+3])
+            r.append(buf[i+3:i+3+l])
+            i = i + 3 + l
+            i_request += 1
+        return r
+
+    def trans_info(self, info_requests):
+        if type(info_requests) == int:  # singleton
+            r = self._trans_info([info_requests])
+            return {info_requests: r[0]}
+        else:
+            results = {}
+            rs = self._trans_info(info_requests)
+            for i in range(len(info_requests)):
+                if info_requests[i] == isc_info_tra_isolation:
+                    v = (b2i(rs[i][0]), b2i(rs[i][1]))
+                else:
+                    v = bytes_to_int(rs[i])
+                results[info_requests[i]] = v
+            return results
+
     def close(self):
         if self.closed:
             return
