@@ -105,6 +105,56 @@ def recv_channel(sock, nbytes, word_alignment=False):
 def send_channel(sock, b):
     sock.send(b)
 
+def params_to_blr(params):
+    "Convert parameter array to BLR and values format."
+    ln = len(params) * 2
+    blr = bytes([5, 2, 4, 0, ln & 255, ln >> 8])
+    values = bytes([])
+    for p in params:
+        t = type(p)
+        if ((sys.version_info[0] == 2 and t == str) or
+            (sys.version_info[0] == 3 and t == bytes)):
+            v = p
+            nbytes = len(v)
+            pad_length = ((4-nbytes) & 3)
+            v += bytes([0]) * pad_length
+            blr += bytes([14, nbytes & 255, nbytes >> 8])
+        elif t == int:
+            v = bint_to_bytes(p, 4)
+            blr += bytes([7, 0])
+        elif t == decimal.Decimal or t == float:
+            if t == float:
+                p = decimal.Decimal(str(p))
+            (sign, digits, exponent) = p.as_tuple()
+            v = 0
+            ln = len(digits)
+            for i in range(ln):
+                v += digits[i] * (10 ** (ln -i-1))
+            if sign:
+                v *= -1
+            v = bint_to_bytes(v, 8)
+            if exponent < 0:
+                exponent += 256
+            blr += bytes([16, exponent])
+        elif t == datetime.date:
+            v = convert_date(p)
+            blr += bytes([12])
+        elif t == datetime.time:
+            v = convert_time(p)
+            blr += bytes([13])
+        elif t == datetime.datetime:
+            v = convert_timestamp(p)
+            blr += bytes([35])
+        elif p == None:
+            v = bytes([0]) * 8
+            blr += bytes([9, 0])
+        values += v
+        blr += bytes([7, 0])
+        values += bytes([0]) * 4 if p != None else bytes([0xff,0xff,0x34,0x8c])
+    blr += bytes([255, 76])    # [blr_end, blr_eoc]
+    return blr, values
+
+
 class WireProtocol:
     buffer_length = 1024
 
@@ -124,6 +174,7 @@ class WireProtocol:
     op_get_segment = 36
     op_close_blob = 39
     op_info_database = 40
+    op_info_transaction = 42
     op_batch_segments = 44
     op_allocate_statement = 62
     op_create_blob2 = 57
@@ -134,6 +185,8 @@ class WireProtocol:
     op_free_statement = 67
     op_prepare_statement = 68
     op_info_sql = 70
+    op_execute2 = 76
+    op_sql_response = 78
     op_drop_database = 81
     op_service_attach = 82
     op_service_detach = 83
@@ -201,6 +254,7 @@ class WireProtocol:
         "convert bytes array to unicode string"
         return b.decode(self.charset_map.get(self.charset, self.charset))
 
+
     @wire_operation
     def _op_connect(self):
         p = xdrlib.Packer()
@@ -371,6 +425,26 @@ class WireProtocol:
         p.pack_int(self.db_handle)
         send_channel(self.sock, p.get_buffer())
 
+    @wire_operation
+    def _op_info_transaction(self, trans_handle, b):
+        p = xdrlib.Packer()
+        p.pack_int(self.op_info_transaction)
+        p.pack_int(trans_handle)
+        p.pack_int(0)
+        p.pack_bytes(b)
+        p.pack_int(self.buffer_length)
+        send_channel(self.sock, p.get_buffer())
+    @wire_operation
+    def _op_info_database(self, b):
+        p = xdrlib.Packer()
+        p.pack_int(self.op_info_database)
+        p.pack_int(self.db_handle)
+        p.pack_int(0)
+        p.pack_bytes(b)
+        p.pack_int(self.buffer_length)
+        send_channel(self.sock, p.get_buffer())
+
+
     @wire_operation
     def _op_free_statement(self, stmt_handle, mode):
         p = xdrlib.Packer()
@@ -415,12 +489,36 @@ class WireProtocol:
             p.pack_int(0)
             send_channel(self.sock, p.get_buffer())
         else:
-            (blr, values) = self.params_to_blr(params)
+            (blr, values) = params_to_blr(params)
             p.pack_bytes(blr)
             p.pack_int(0)
             p.pack_int(1)
             send_channel(self.sock, p.get_buffer() + values)
 
+    @wire_operation
+    def _op_execute2(self, stmt_handle, trans_handle, params, output_blr):
+        p = xdrlib.Packer()
+        p.pack_int(self.op_execute2)
+        p.pack_int(stmt_handle)
+        p.pack_int(trans_handle)
+
+        if len(params) == 0:
+            p.pack_bytes(bytes([]))
+            p.pack_int(0)
+            p.pack_int(0)
+            send_channel(self.sock, p.get_buffer())
+        else:
+            (blr, values) = params_to_blr(params)
+            p.pack_bytes(blr)
+            p.pack_int(0)
+            p.pack_int(1)
+            send_channel(self.sock, p.get_buffer() + values)
+
+        p = xdrlib.Packer()
+        p.pack_bytes(output_blr)
+        p.pack_int(0)
+        send_channel(self.sock, p.get_buffer())
+
     @wire_operation
     def _op_execute_immediate(self, trans_handle, db_handle, sql='', params=[],
                                     in_msg='', out_msg='', possible_requests=0):
@@ -435,7 +533,7 @@ class WireProtocol:
             r += bint_to_bytes(0, 2)    # in_blr len
             values = bytes([])
         else:
-            (blr, values) = self.params_to_blr(params)
+            (blr, values) = params_to_blr(params)
             r += bint_to_bytes(len(blr), 2) + blr
         r += bint_to_bytes(len(in_msg), 2) + in_msg
         r += bint_to_bytes(0, 2)    # unknown short int 0
@@ -492,18 +590,18 @@ class WireProtocol:
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_open_blob(self, blob_id):
+    def _op_open_blob(self, blob_id, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_open_blob)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
         send_channel(self.sock, p.get_buffer() + blob_id)
 
     @wire_operation
-    def _op_create_blob2(self):
+    def _op_create_blob2(self, trans_handle):
         p = xdrlib.Packer()
         p.pack_int(self.op_create_blob2)
         p.pack_int(0)
-        p.pack_int(self.trans_handle)
+        p.pack_int(trans_handle)
         p.pack_int(0)
         p.pack_int(0)
         send_channel(self.sock, p.get_buffer())
@@ -579,3 +677,31 @@ class WireProtocol:
 
         return (h, oid, buf)
 
+    @wire_operation
+    def _op_sql_response(self, xsqlda):
+        b = recv_channel(self.sock, 4)
+        while bytes_to_bint(b) == self.op_dummy:
+            b = recv_channel(self.sock, 4)
+        if bytes_to_bint(b) != self.op_sql_response:
+            raise InternalError
+
+        b = recv_channel(self.sock, 4)
+        count = bytes_to_bint(b[:4])
+
+        r = []
+        for i in range(len(xsqlda)):
+            x = xsqlda[i]
+            if x.io_length() < 0:
+                b = recv_channel(self.sock, 4)
+                ln = bytes_to_bint(b)
+            else:
+                ln = x.io_length()
+            raw_value = recv_channel(self.sock, ln, True)
+            if recv_channel(self.sock, 4) == bytes([0]) * 4: # Not NULL
+                r.append(x.value(raw_value))
+            else:
+                r.append(None)
+
+        b = recv_channel(self.sock, 32)     # ??? why 32 bytes skip
+
+        return r
